"""
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
from typing import List
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    makeDirPath,
                                    makeMap,
                                    readDataFile,
                                    successiveParents)
from drapi.code.drapi.deIdentificationFunctions import functionFromSettings
# Super-local
from common import (IRB_NUMBER,
                    NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA,
                    VARIABLE_ALIASES,
                    VARIABLE_SUFFIXES)


# Functions
def convertColumnsHash(listOfPortionDirs: list,
                       listOfPortionNames: List[str],
                       LIST_OF_PORTION_CONDITIONS: list,
                       IRB_NUMBER: int,
                       VARIABLE_SUFFIXES: dict,
                       logger: logging.Logger,
                       deIdentificationFunctions: dict,
                       VARIABLE_ALIASES: dict,
                       CHUNKSIZE: int,
                       SCRIPT_TEST_MODE: bool = False):
    """
    """
    for variableName, func in deIdentificationFunctions.items():
        logger.info(f"""  {variableName}: {func(0)}.""")

    # Assert all portions have file criteria accounted for
    if len(listOfPortionDirs) == len(LIST_OF_PORTION_CONDITIONS):
        pass
    else:
        raise Exception("The number of portions does not equal the number of portion conditions.")

    # Collect columns to convert
    columnsToConvert = sorted(list(deIdentificationFunctions.keys()))

    # Load de-identification maps for each variable that needs to be de-identified
    logger.info("""Loading de-identification maps for each variable that needs to be de-identified.""")
    mapsColumnNames = {}
    variablesCollected = [fname for fname in deIdentificationFunctions.keys()]
    for varName in variablesCollected:
        logger.info(f"""  Creating de-identified variable name for "{varName}".""")
        mapsColumnNames[varName] = f"De-identified {varName}"
    # Add aliases to `mapsColumnNames`
    for varName in VARIABLE_ALIASES.keys():
        varAlias = VARIABLE_ALIASES[varName]
        map_ = makeMap(IDset=set(),
                       IDName=varName,
                       startFrom=1,
                       irbNumber=IRB_NUMBER,
                       suffix=VARIABLE_SUFFIXES[varAlias]["deIdIDSuffix"],
                       columnSuffix=varName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        mapsColumnNames[varName] = map_.columns[-1]

    # QA: Test de-identification functions
    logger.info("""QA: Testing de-identification functions.""")
    for variableName, func in DE_IDENTIFICATION_FUNCTIONS.items():
        logger.info(f"""  {variableName}: {func(1)}.""")

    # De-identify columns
    logger.info("""De-identifying files.""")
    allowLogging = False
    # Set file options: Map file
    for portionName, directory, fileConditions in zip(listOfPortionNames, listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        directory = Path(directory)  # For type hinting
        # Act on directory
        logger.info(f"""Working on directory "{choosePathToLog(directory, rootDirectory)}".""")
        listOfFiles = sorted(list(directory.iterdir()))
        for file in listOfFiles:
            logger.info(f"""  Working on file "{choosePathToLog(file, rootDirectory)}".""")
            fileOptions = {columnName: {"header": True,
                                        "mode": "w"} for columnName in columnsToConvert}
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options: Data file
                exportPath = runOutputDir.joinpath("Portions", portionName, f"{file.stem}.CSV")
                makeDirPath(exportPath.parent)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                logger.info("""  ..  Reading file to count the number of chunks.""")
                # >>> TEST BLOCK >>>  # TODO Remove
                if SCRIPT_TEST_MODE:
                    it1Total = 100
                else:
                    it1Total = sum([1 for _ in readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)])
                # <<< TEST BLOCK <<<
                logger.info(f"""  ..  This file has {it1Total:,} chunks.""")
                # Calculate logging requency
                if it1Total < MESSAGE_MODULO_CHUNKS:
                    moduloChunks = it1Total
                else:
                    moduloChunks = round(it1Total / MESSAGE_MODULO_CHUNKS)
                if it1Total / moduloChunks < 100:
                    moduloChunks = 1
                else:
                    pass
                dfChunks = readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)
                filePreview = readDataFile(file, nrows=2)
                filePreview = pd.DataFrame(filePreview)
                for it1, dfChunk0 in enumerate(dfChunks, start=1):
                    # >>> TEST BLOCK >>>  # TODO Remove
                    if SCRIPT_TEST_MODE:
                        if it1 > 1:
                            logger.info("""  ..  `SCRIPT_TEST_MODE` engaged. BREAKING.""")
                            break
                    # <<< TEST BLOCK <<<
                    dfChunk = pd.DataFrame(dfChunk0)
                    # Work on chunk
                    if it1 % moduloChunks == 0:
                        allowLogging = True
                    else:
                        allowLogging = False
                    if allowLogging:
                        logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        if columnName in VARIABLE_ALIASES.keys():
                            columnNameAlias = VARIABLE_ALIASES[columnName]
                            hasAlias = True
                        else:
                            columnNameAlias = columnName
                            hasAlias = False
                        # Logging block
                        if allowLogging:
                            if hasAlias:
                                logger.info(f"""  ..    Working on column "{columnName}" aliased as "{columnNameAlias}".""")
                            else:
                                logger.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in columnsToConvert or columnNameAlias in columnsToConvert:
                            if allowLogging:
                                logger.info("""  ..  ..  Column must be converted. Converting column...""")
                            # De-identify column: create de-identified values
                            newColumnName = mapsColumnNames[columnName]
                            newColumn = dfChunk[columnName].apply(deIdentificationFunctions[columnNameAlias])
                            newColumn = pd.Series(newColumn)  # For type hinting
                            newColumnWithOldName = newColumn
                            newColumnWithNewName = newColumn.rename(newColumnName)  # We can change `newColumnName` with the aliased version
                            # QA: Column/Variable Map: Variables
                            mapSavePath = runOutputDir.joinpath("Metadata", "Maps by Portion", portionName, f"{columnNameAlias}", f"{file.stem}.CSV")
                            makeDirPath(mapSavePath.parent)
                            mapHeader = fileOptions[columnNameAlias]["header"]
                            mapMode = fileOptions[columnNameAlias]["mode"]
                            # QA: Column/Variable Map: Save mapped column
                            chunkColumnMap = pd.concat([dfChunk[columnName], newColumnWithNewName], axis=1).drop_duplicates()
                            logger.info(f"""  ..  ..  Saving QA table to "{choosePathToLog(mapSavePath.absolute(), rootPath=rootDirectory)}".""")
                            logger.info(f"""  ..  ..  Saving QA table to `mapHeader` set to "{mapHeader}".""")
                            logger.info(f"""  ..  ..  Saving QA table to `mapMode` set to "{mapMode}".""")
                            chunkColumnMap.to_csv(mapSavePath, index=False, header=mapHeader, mode=mapMode)
                            fileOptions[columnNameAlias]["header"] = False
                            fileOptions[columnNameAlias]["mode"] = "a"
                            # De-identify column: Replace old values
                            dfChunk[columnName] = newColumnWithOldName
                            dfChunk = dfChunk.rename(columns={columnName: newColumnName})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileHeaders = False
                    fileMode = "a"
                    if allowLogging:
                        logger.info(f"""  ..  Chunk saved to "{choosePathToLog(exportPath, rootDirectory)}".""")
                # End file lopp
                pass
            else:
                logger.info("""    This file does not need to be processed.""")


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main: Multiple variable option: Dictionary format
    parser.add_argument("--DICTIONARY_OF_MAPPING_ARGUMENTS",
                        type=json.loads,
                        help="""The input must be of this format: {variableName: {"Encryption Type": encryptionType, "Encryption Secret": encryptionSecret}}""")

    # Arguments: Main: Multiple variable option: String format
    # NOTE TODO Not implemented yet. This should allow the user to input a space-delimitted list of arguments, easier than the dictionary option
    parser.add_argument("--MAPPING_ARGUMENTS",
                        nargs="+",
                        type=json.loads,
                        help="""The input must be of this format: {variableName: {"Encryption Type": encryptionType, "Encryption Secret": encryptionSecret}}""")

    # Arguments: Main: Single variable option
    parser.add_argument("--VARIABLE_NAME_TO_ENCRYPT",
                        type=str,
                        help="")
    helptext = r"""1: Additive encryption. E.g., `encryptValue1(value='123456789', secret=1)  # 123456790`.
2: Encrypt with character-wise XOR operation of both operands, with the second operand rotating over the set of character-wise values in `secretkey`. E.g., `encryptValue1(value='123456789', secret='password')  # 'AS@GBYE\I'
3: Encrypt with whole-value XOR operation. Requires both operands to be integers. E.g., `encryptValue1(value=123456789, secret=111111111)  # 1326016938`"""
    parser.add_argument("--ENCRYPTION_TYPE",
                        type=int,
                        choices=[1, 2, 3],
                        help=helptext)
    parser.add_argument("--ENCRYPTION_SECRET",
                        type=lambda stringValue: int(stringValue) if stringValue.isnumeric() else stringValue,
                        help="We expect an integer or stringValue. If the stringValue is purely numbers, it will be converted to an integer object. If you don't pass an argument then a secret value will be generated for you at random.")

    # Arguments: Main
    parser.add_argument("SCRIPT_TEST_MODE",
                        type=lambda stringValue: True if stringValue.lower() == "true" else False if stringValue.lower() == "false" else None,
                        help=""" Choose one of {{True, False}}""")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    DICTIONARY_OF_MAPPING_ARGUMENTS = argNamespace.DICTIONARY_OF_MAPPING_ARGUMENTS
    MAPPING_ARGUMENTS = argNamespace.MAPPING_ARGUMENTS

    VARIABLE_NAME_TO_ENCRYPT = argNamespace.VARIABLE_NAME_TO_ENCRYPT
    ENCRYPTION_TYPE = argNamespace.ENCRYPTION_TYPE
    ENCRYPTION_SECRET = argNamespace.ENCRYPTION_SECRET

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # Argument parsing: Additional checks  # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    if DICTIONARY_OF_MAPPING_ARGUMENTS and (VARIABLE_NAME_TO_ENCRYPT or ENCRYPTION_TYPE or ENCRYPTION_SECRET):
        parser.error("""This program is meant to function one of two ways. Either
1. Pass `DICTIONARY_OF_MAPPING_ARGUMENTS`, or
2. Pass each of
    2. a. `VARIABLE_NAME_TO_ENCRYPT`
    2. b. `ENCRYPTION_TYPE`
    2. c. `ENCRYPTION_SECRET`""")

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Meta
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    # >>> Argument prep >>>

    LIST_OF_PORTION_NAMES = ["OMOP",
                             "Notes Text",
                             "Notes Metadata"]
    LIST_OF_PORTION_DIRS = ["../Concatenated Results/data/output/convertColumnsGeneral/2024-02-23 18-11-43",
                            "/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Notes/original_note_pull/data/Output/free_text/cohort_note/tsv",
                            "/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Notes/original_note_pull/data/Output/free_text"]
    LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA,
                                  [lambda pathObj: pathObj.suffix.lower() == ".tsv"],
                                  NOTES_PORTION_FILE_CRITERIA]  # TODO Remove hard-coded option

    # Conform mapping arguments
    if DICTIONARY_OF_MAPPING_ARGUMENTS:
        mappingArguments = [{variableName: tu} for variableName, tu in DICTIONARY_OF_MAPPING_ARGUMENTS.items()]
    elif MAPPING_ARGUMENTS:
        mappingArguments = [di for di in MAPPING_ARGUMENTS]
    elif VARIABLE_NAME_TO_ENCRYPT and ENCRYPTION_TYPE and ENCRYPTION_SECRET:
        mappingArguments = [{VARIABLE_NAME_TO_ENCRYPT: [ENCRYPTION_TYPE, ENCRYPTION_SECRET]}]
    else:
        raise Exception("This should not happen")

    # Define the de-identification functions for each variable.
    DE_IDENTIFICATION_FUNCTIONS = {}
    mappingSettings = {}
    for di in mappingArguments:
        variableName = list(di.keys())[0]
        encryptionType, encryptionSecret0 = list(di.values())[0]
        encryptionSecret, variableFunction = functionFromSettings(ENCRYPTION_TYPE=encryptionType,
                                                                  ENCRYPTION_SECRET=encryptionSecret0,
                                                                  IRB_NUMBER=IRB_NUMBER)
        DE_IDENTIFICATION_FUNCTIONS[variableName] = lambda value: variableFunction(value=value,
                                                                                   suffix=VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"],
                                                                                   secret=encryptionSecret)
        mappingSettings[variableName] = {"Encryption Type": encryptionType,
                                         "Encryption Secret (Input)": encryptionSecret0,
                                         "Encryption Secret (Final)": encryptionSecret}

    # QA: Test de-identification functions
    logger.info("""QA: Testing de-identification functions.""")
    for variableName, func in DE_IDENTIFICATION_FUNCTIONS.items():
        logger.info(f"""  {variableName}: {func(1)}.""")

    # <<< Argument prep <<<

    # Save secrets
    secretPath = runOutputDir.joinpath("Metadata", "Mapping Settings", "Mapping Settings.CSV")
    makeDirPath(secretPath.parent)
    df = pd.DataFrame.from_dict(data=mappingSettings, orient="index")
    df = df.sort_index()
    df.to_csv(secretPath)

    convertColumnsHash(listOfPortionDirs=LIST_OF_PORTION_DIRS,
                       LIST_OF_PORTION_CONDITIONS=LIST_OF_PORTION_CONDITIONS,
                       listOfPortionNames=LIST_OF_PORTION_NAMES,
                       deIdentificationFunctions=DE_IDENTIFICATION_FUNCTIONS,
                       logger=logger,
                       IRB_NUMBER=IRB_NUMBER,
                       VARIABLE_ALIASES=VARIABLE_ALIASES,
                       VARIABLE_SUFFIXES=VARIABLE_SUFFIXES,
                       CHUNKSIZE=CHUNKSIZE,
                       SCRIPT_TEST_MODE=SCRIPT_TEST_MODE)

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Iterates over the files that would be processed in the pipeline and runs quality tests on them. Currently the quality tests are

  1. Check for delimmeter issues. This is done by reading the whole file with `pd.read_csv`. Usually if there's an unexpected presence or absence of a delimmeter this will raise an error.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""

__all__ = ["runIntermediateDataDir"]

import logging
import sys
from pathlib import Path
# Third-party packages
from pandas.errors import ParserError
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    successiveParents)
from common import (DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    BO_PORTION_DIR_MAC,
                    BO_PORTION_DIR_WIN,
                    BO_PORTION_FILE_CRITERIA)

# Arguments
CHUNK_SIZE = 50000

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Arguments: Portion Paths and conditions
MAC_PATHS = [BO_PORTION_DIR_MAC]
WIN_PATHS = [BO_PORTION_DIR_WIN]

LIST_OF_PORTION_CONDITIONS = [BO_PORTION_FILE_CRITERIA]

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Data quality check
    logging.info("""Getting the set of values for each variable to de-identify.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logging.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logging.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Read file
                logging.info("""    This file has met all conditions for testing.""")
                # Test 1: Make sure all lines have the same number of delimiters
                logging.info("""  ..  Test 1: Make sure all lines have the same number of delimiters.""")
                try:
                    numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE)])
                    logging.info("""  ..    There are no apparent problems reading this file.""")
                except ParserError as err:
                    args = err
                    logging.info("""  ..    This file raised an error: "{err}".""")
                # Test 2: ...
                pass
            else:
                logging.info("""    This file does not need to be tested.""")

    # Return path to sets fo ID values
    # TODO If this is implemented as a function, instead of a stand-alone script, return `runOutputDir` to define `setsPathDir` in the "makeMap" scripts.
    logging.info(f"""Finished collecting the set of ID values to de-identify. The set files are located in "{runOutputDir.relative_to(projectDir)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Converts columns by using one of three pre-defined mapping functions.
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
# Third-party packages
import pandas as pd
import pathos.pools as pp
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.convertColumns.hash import convertColumnsHash
from drapi.code.drapi.deIdentificationFunctions import functionFromSettings
from drapi.code.drapi.parseAliasArguments import parseAliasArguments
from drapi.code.drapi.constants.phiVariables import VARIABLE_SUFFIXES
from drapi.code.drapi.constants.variableAliases import VARIABLE_ALIASES


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()
    # Arguments: Main
    parser.add_argument("IRB_NUMBER",
                        type=str,
                        help="""The IRB number, usually in the form "<PROTOCOL_ABBREVIATION><SERIAL_NUMBER>, where <PROTOCOL_ABBREVIATION> is something like "IRB", "NH", "WIRB", and <SERIAL_NUMBER> is a 9-digit number where the first four digits is the four-digit year, e.g., "2024".""")
    parser.add_argument("SCRIPT_TEST_MODE",
                        type=lambda stringValue: True if stringValue.lower() == "true" else False if stringValue.lower() == "false" else None,
                        help=""" Choose one of {{True, False}}""")


    # Arguments: Main: Multiple variable option: Dictionary format
    parser.add_argument("--DICTIONARY_OF_MAPPING_ARGUMENTS",
                        type=json.loads,
                        help="""The input must be of this format: {variableName: {"Encryption Type": encryptionType, "Encryption Secret": encryptionSecret}}""")

    # Arguments: Main: Multiple variable option: String format
    # NOTE TODO Not implemented yet. This should allow the user to input a space-delimitted list of arguments, easier than the dictionary option
    parser.add_argument("--MAPPING_ARGUMENTS",
                        nargs="+",
                        type=json.loads,
                        help="""The input must be of this format: {variableName: {"Encryption Type": encryptionType, "Encryption Secret": encryptionSecret}}""")
    
    # Arguments: Main: File paths and portion names
    parser.add_argument("PATHS",
                        nargs="+",
                        type=str,
                        help="")
    parser.add_argument("--PORTION_NAME",
                        type=str,
                        help="""The name of the set or subset of data, usually one of "BO", "Clinical Text", "Clinical Text Metadata", "Line-level", "OMOP", "SDOH".""")
    parser.add_argument("--BY",
                        choices=["dir", "file", "file-dir"],
                        help="""Whether the `PATHS` values passed are for directories or files. If you are passing directories but wish their contents to be processed in parallel, use the option "file-dir".""")
    

    # Arguments: Main: Single variable option
    parser.add_argument("--VARIABLE_NAME_TO_ENCRYPT",
                        type=str,
                        help="")
    helptext = r"""1: Additive encryption. E.g., `encryptValue1(value='123456789', secret=1)  # 123456790`.
2: Encrypt with character-wise XOR operation of both operands, with the second operand rotating over the set of character-wise values in `secretkey`. E.g., `encryptValue1(value='123456789', secret='password')  # 'AS@GBYE\I'
3: Encrypt with whole-value XOR operation. Requires both operands to be integers. E.g., `encryptValue1(value=123456789, secret=111111111)  # 1326016938`"""
    parser.add_argument("--ENCRYPTION_TYPE",
                        type=int,
                        choices=[1, 2, 3],
                        help=helptext)
    parser.add_argument("--ENCRYPTION_SECRET",
                        type=lambda stringValue: int(stringValue) if stringValue.isnumeric() else stringValue,
                        help="We expect an integer or stringValue. If the stringValue is purely numbers, it will be converted to an integer object. If you don't pass an argument then a secret value will be generated for you at random.")

    parser.add_argument("--CUSTOM_ALIASES",
                        help="""A JSON-formatted string of the form {`ALIAS`: `VARIABLE_NAME`}, where `VARIABLE_NAME` is the BO version of a variable name, and `ALIAS` is an alias of the variable name. An example is {"EncounterCSN": "Encounter # (CSN)"}.""",
                        type=json.loads)
    parser.add_argument("--DEFAULT_ALIASES",
                        help="""Indicates whether to include the default IDR aliases for variables.""",
                        action="store_true")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    IRB_NUMBER = argNamespace.IRB_NUMBER

    DICTIONARY_OF_MAPPING_ARGUMENTS = argNamespace.DICTIONARY_OF_MAPPING_ARGUMENTS
    MAPPING_ARGUMENTS = argNamespace.MAPPING_ARGUMENTS

    VARIABLE_NAME_TO_ENCRYPT = argNamespace.VARIABLE_NAME_TO_ENCRYPT
    ENCRYPTION_TYPE = argNamespace.ENCRYPTION_TYPE
    ENCRYPTION_SECRET = argNamespace.ENCRYPTION_SECRET

    BY = argNamespace.BY
    PATHS = argNamespace.PATHS
    PORTION_NAME = argNamespace.PORTION_NAME

    CUSTOM_ALIASES = argNamespace.CUSTOM_ALIASES
    DEFAULT_ALIASES = argNamespace.DEFAULT_ALIASES

    # TODO Proper definition
    listOfPortionDirs, listOfPortionNames, LIST_OF_PORTION_CONDITIONS = None, None, None

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Custom argument parsing: aliases
    variableAliases = parseAliasArguments(customAliases=CUSTOM_ALIASES,
                                          useDefaultAliases=DEFAULT_ALIASES,
                                          defaultAliases=VARIABLE_ALIASES)
    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # Argument parsing: Additional checks  # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    if DICTIONARY_OF_MAPPING_ARGUMENTS and (VARIABLE_NAME_TO_ENCRYPT or ENCRYPTION_TYPE or ENCRYPTION_SECRET):
        parser.error("""This program is meant to function one of two ways. Either
1. Pass `DICTIONARY_OF_MAPPING_ARGUMENTS`, or
2. Pass each of
    2. a. `VARIABLE_NAME_TO_ENCRYPT`
    2. b. `ENCRYPTION_TYPE`
    2. c. `ENCRYPTION_SECRET`""")

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Meta
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    # >>> Argument prep >>>
    # Conform mapping arguments
    if DICTIONARY_OF_MAPPING_ARGUMENTS:
        mappingArguments = [{variableName: tu} for variableName, tu in DICTIONARY_OF_MAPPING_ARGUMENTS.items()]
    elif MAPPING_ARGUMENTS:
        mappingArguments = [di for di in MAPPING_ARGUMENTS]
    elif VARIABLE_NAME_TO_ENCRYPT and ENCRYPTION_TYPE and ENCRYPTION_SECRET:
        mappingArguments = [{VARIABLE_NAME_TO_ENCRYPT: [ENCRYPTION_TYPE, ENCRYPTION_SECRET]}]
    else:
        raise Exception("This should not happen")

    # Define the de-identification functions for each variable.
    deIdentificationFunctions = {}
    mappingSettings = {}
    for di in mappingArguments:
        variableName = list(di.keys())[0]
        encryptionType, encryptionSecret0 = list(di.values())[0]
        encryptionSecret, variableFunction = functionFromSettings(ENCRYPTION_TYPE=encryptionType,
                                                                  ENCRYPTION_SECRET=encryptionSecret0,
                                                                  IRB_NUMBER=IRB_NUMBER,
                                                                  suffix=VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"])
        deIdentificationFunctions[variableName] = variableFunction
        mappingSettings[variableName] = {"Encryption Type": encryptionType,
                                         "Encryption Secret (Input)": encryptionSecret0,
                                         "Encryption Secret (Final)": encryptionSecret}

    # QA: Test de-identification functions
    logger.info("""QA: Testing de-identification functions.""")
    for variableName, func in deIdentificationFunctions.items():
        logger.info(f"""  {variableName}: {func(1)}.""")

    PARGS = []
    if BY.lower() == "dir":
        zipObj = zip(LIST_OF_PORTION_CONDITIONS,
                     listOfPortionDirs,
                     listOfPortionNames)
        for portionCondition, portionDir, portionName in zipObj:
            argsClone = (BY,
                         # Arguments used in preparation of case functions
                         IRB_NUMBER,
                         VARIABLE_SUFFIXES,
                         # Arguments used in preparation AND passed to other functions: Common to both cases
                         VARIABLE_ALIASES,
                         deIdentificationFunctions,
                         logger,
                         # Arguments passed to other functions: Common to both cases
                         rootDirectory,
                         runOutputDir,
                         CHUNKSIZE,
                         MESSAGE_MODULO_CHUNKS,
                         SCRIPT_TEST_MODE,
                         # Arguments passed to other functions: "dir" case
                         portionDir,
                         portionCondition,
                         portionName,
                         # Arguments passed to other functions: "file" case
                         None,
                         None)
            PARGS.append(argsClone)
    # Arguments passed to other functions: "file" case
    elif BY.lower() == "file":
        li = list(map(lambda fpath: (PORTION_NAME, fpath), PATHS))
        for portionName, fpath in li:
            argsClone = (BY,
                         # Arguments used in preparation of case functions
                         IRB_NUMBER,
                         VARIABLE_SUFFIXES,
                         # Arguments used in preparation AND passed to other functions: Common to both cases
                         VARIABLE_ALIASES,
                         deIdentificationFunctions,
                         logger,
                         # Arguments passed to other functions: Common to both cases
                         rootDirectory,
                         runOutputDir,
                         CHUNKSIZE,
                         MESSAGE_MODULO_CHUNKS,
                         SCRIPT_TEST_MODE,
                         # Arguments passed to other functions: "dir" case
                         None,
                         None,
                         None,
                         # Arguments passed to other functions: "file" case
                         fpath,
                         portionName)
            PARGS.append(argsClone)
    elif BY.lower() == "file-dir":
        paths = []
        for pathString in PATHS:
            dpath = Path(pathString)
            paths.extend(dpath.iterdir())
        li = list(map(lambda fpath: (PORTION_NAME, fpath), paths))
        for portionName, fpath in li:
            argsClone = (BY,
                         # Arguments used in preparation of case functions
                         IRB_NUMBER,
                         VARIABLE_SUFFIXES,
                         # Arguments used in preparation AND passed to other functions: Common to both cases
                         VARIABLE_ALIASES,
                         deIdentificationFunctions,
                         logger,
                         # Arguments passed to other functions: Common to both cases
                         rootDirectory,
                         runOutputDir,
                         CHUNKSIZE,
                         MESSAGE_MODULO_CHUNKS,
                         SCRIPT_TEST_MODE,
                         # Arguments passed to other functions: "dir" case
                         None,
                         None,
                         None,
                         # Arguments passed to other functions: "file" case
                         fpath,
                         portionName)
            PARGS.append(argsClone)
    else:
        raise Exception("This should not happen!")
    parallelizationArgListString = pprint.pformat(PARGS)
    logger.info(f"""`PARGS`:\n{parallelizationArgListString}""")

    # <<< Argument prep <<<

    # Save secrets
    secretPath = runOutputDir.joinpath("Metadata", "Mapping Settings", "Mapping Settings.CSV")
    makeDirPath(secretPath.parent)
    df = pd.DataFrame.from_dict(data=mappingSettings, orient="index")
    df = df.sort_index()
    df.to_csv(secretPath)

    with pp._ProcessPool() as pool:
        results = pool.starmap(convertColumnsHash, PARGS)

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Converts a map CSV to a SQLite database.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
import sqlite3
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    successiveParents)
from drapi.code.drapi.constants.constants import DATA_TYPES_DICT_SQL
# Super-local
pass

# Arguments
if True:
    MAP_PATH = r"/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Intermediate Results/SQL Portion/data/output/getData/2024-02-22 16-44-29"  # String, the path to the map CSV file or the directory that contains a series of files that combined form a map CSV.
    VARIABLE_NAME_FROM = "provider_id"
    VARIABLE_NAME_TO = "Provider Key"
elif True:
    MAP_PATH = r"..\Intermediate Results\SQL Portion\data\output\getData\2024-01-19 21-42-27"  # String, the path to the map CSV file or the directory that contains a series of files that combined form a map CSV.
    VARIABLE_NAME_FROM = "visit_occurrence_id"
    VARIABLE_NAME_TO = "Encounter # (CSN)"
elif True:
    MAP_PATH = r"..\OMOP_Structured_Data\data\output\mapping\MRN_mapping.csv"  # String, the path to the map CSV file or the directory that contains a series of files that combined form a map CSV.
    VARIABLE_NAME_FROM = "person_id"
    VARIABLE_NAME_TO = "PatientKey"

CHUNKSIZE = 50000  # The number of rows to read at a time from the CSV using Pandas `chunksize`
MESSAGE_MODULO_CHUNKS = 50  # How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`
MESSAGE_MODULO_FILES = 100  # How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 4

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
# ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
#  "IRB_DIRECTORY",                 # noqa
#  "DATA_REQUEST_DIRECTORY",        # noqa
#  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Functions


def createSqliteDatabase(databasePath: str,
                         variableNameFrom: str,
                         variableNameTo: str,
                         dataTypesDictSql: dict) -> None:
    """
    """
    with sqlite3.connect(databasePath) as connection:
        cursor = connection.cursor()
        cursor.execute(f"""CREATE TABLE '{tableName}'
                    (
                    '{variableNameFrom}' {dataTypesDictSql[variableNameFrom]},
                    '{variableNameTo}' {dataTypesDictSql[variableNameTo]}
                    )""")


def insertIntoSqliteDatabase(databasePath: str,
                             variableNameFrom: str,
                             variableNameTo: str,
                             dataFrame: pd.DataFrame) -> None:
    """
    """
    with sqlite3.connect(databasePath) as connection:
        cursor = connection.cursor()
        for it, (label, rowSeries) in enumerate(dataFrame.iterrows(), start=1):
            cursor.execute(f"""INSERT INTO '{tableName}'
                        (
                        '{variableNameFrom}',
                        '{variableNameTo}'
                        )
                        values
                        (
                        {rowSeries[variableNameFrom]},
                        {rowSeries[variableNameTo]}
                        )""")
        connection.commit()


# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `MAP_PATH`: "{MAP_PATH}"
    `VARIABLE_NAME_FROM`: "{VARIABLE_NAME_FROM}"
    `VARIABLE_NAME_TO`: "{VARIABLE_NAME_TO}"
    `CHUNKSIZE`: "{CHUNKSIZE}"
    `MESSAGE_MODULO_CHUNKS`: "{MESSAGE_MODULO_CHUNKS}"
    `MESSAGE_MODULO_FILES`: "{MESSAGE_MODULO_FILES}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Create script variables
    mapPath = Path(MAP_PATH)
    databaseName = f"""{VARIABLE_NAME_FROM} to {VARIABLE_NAME_TO}"""
    tableName = databaseName[:]
    databasePath = runOutputDir.joinpath(f"{databaseName}.db")

    # Create SQLite database
    logger.info("""Creating SQLite database.""")
    if mapPath.is_file():
        logger.info("""  The path provided is for a file.""")
        logger.info("""  Counting the number of chunks in the file.""")
        fileChunks = readDataFile(fname=mapPath,
                                  chunksize=CHUNKSIZE)
        numChunks = sum([1 for _ in fileChunks])
        logger.info("""  Counting the number of chunks in the file - done.""")
        fileChunks = readDataFile(fname=mapPath,
                                  chunksize=CHUNKSIZE)
        if numChunks < MESSAGE_MODULO_CHUNKS:
            moduloChunks = numChunks
        else:
            moduloChunks = round(numChunks / MESSAGE_MODULO_CHUNKS)
        createSqliteDatabase(databasePath=databasePath,
                             variableNameFrom=VARIABLE_NAME_FROM,
                             variableNameTo=VARIABLE_NAME_TO,
                             dataTypesDictSql=DATA_TYPES_DICT_SQL)
        for it1, dfChunk in enumerate(fileChunks, start=1):
            if it1 == 1 or it1 % moduloChunks == 0:
                logger.info(f"""    Working on chunk {it1:,} of {numChunks:,}.""")
            insertIntoSqliteDatabase(databasePath=databasePath,
                                     variableNameFrom=VARIABLE_NAME_FROM,
                                     variableNameTo=VARIABLE_NAME_TO,
                                     dataFrame=dfChunk)
    elif mapPath.is_dir():
        logger.info("""  The path provided is for a directory.""")
        logger.info("""  Counting the number of files in the directory.""")
        files = sorted(list(mapPath.iterdir()))
        numFiles = len(files)
        logger.info("""  Counting the number of files in the directory - done.""")
        if numFiles < MESSAGE_MODULO_FILES:
            moduloFiles = numFiles
        else:
            moduloFiles = round(numFiles / MESSAGE_MODULO_FILES)
        createSqliteDatabase(databasePath=databasePath,
                             variableNameFrom=VARIABLE_NAME_FROM,
                             variableNameTo=VARIABLE_NAME_TO,
                             dataTypesDictSql=DATA_TYPES_DICT_SQL)
        for it2, fpath in enumerate(files, start=1):
            if it2 == 1 or it2 % moduloFiles == 0:
                allowLogs = True
            else:
                allowLogs = False
            if allowLogs:
                logger.info(f"""  Working on file {it2:,} of {numFiles:,}.""")
                logger.info("""  Counting the number of chunks in the file.""")
            fileChunks = readDataFile(fname=fpath,
                                      chunksize=CHUNKSIZE)
            numChunks = sum([1 for _ in fileChunks])
            if allowLogs:
                logger.info("""  Counting the number of chunks in the file - done.""")
            fileChunks = readDataFile(fname=fpath,
                                      chunksize=CHUNKSIZE)
            if numChunks < MESSAGE_MODULO_CHUNKS:
                moduloChunks = numChunks
            else:
                moduloChunks = round(numChunks / MESSAGE_MODULO_CHUNKS)
            for it1, dfChunk in enumerate(fileChunks, start=1):
                if allowLogs:
                    if it1 == 1 or it1 % moduloChunks == 0:
                        logger.info(f"""    Working on chunk {it1:,} of {numChunks:,}.""")
                insertIntoSqliteDatabase(databasePath=databasePath,
                                         variableNameFrom=VARIABLE_NAME_FROM,
                                         variableNameTo=VARIABLE_NAME_TO,
                                         dataFrame=dfChunk)

    else:
        raise Exception(f"""Input file "{mapPath.absolute().relative_to(rootDirectory)}" is neither a file nor a directory.""")
    logger.info("""Creating SQLite database - done.""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Makes de-identification maps, building on existing maps.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""

import logging
import re
from collections import OrderedDict
from pathlib import Path
# Third-party packages
import pandas as pd
import pprint
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    getPercentDifference,
                                    successiveParents,
                                    makeMap)
from drapi.code.drapi.constants.phiVariables import (VARIABLE_NAME_TO_FILE_NAME_DICT,
                                          FILE_NAME_TO_VARIABLE_NAME_DICT)
# Project parameters: General
from common import (IRB_NUMBER,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    OLD_MAPS_DIR_PATH,
                    VARIABLE_ALIASES,
                    VARIABLE_SUFFIXES)

# Arguments
NEW_MAPS_DIR_PATH = Path(r"..\Concatenated Results\data\output\makeMapsFromOthers\...")  # TODO
DE_IDENTIFICATION_MAP_STYLE = "lemur"

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Map new maps to variable names
    logging.info("""Mapping new maps to variable names.""")
    pattern = r"^([a-zA-Z_0-9\(\)# -]+) map"
    newMapsFileDict = {}
    for fpath in NEW_MAPS_DIR_PATH.iterdir():
        fname = fpath.stem
        obj = re.match(pattern, fname)
        if obj:
            extractedText = obj.groups()[0]
            variableName = FILE_NAME_TO_VARIABLE_NAME_DICT[extractedText]
        else:
            msg = "File name is of an unexpected format."
            logging.info(msg)
            raise Exception(msg)
        newMapsFileDict[variableName] = [fpath]
    logging.info("The list of variables extracted from the new maps:")
    for el in sorted(newMapsFileDict.keys()):
        logging.info(f"  {el}")

    # Match old and new maps by variable name
    logging.info("""Matching old and new maps by variable name.""")
    matchedMaps0 = {}
    MAPS_COLLECTION_DICT = {"Old Maps": OLD_MAPS_DIR_PATH,
                            "New Maps": newMapsFileDict}
    for mapCollectionName, mapCollection in MAPS_COLLECTION_DICT.items():
        logging.info(f"""  Working on map collection "{mapCollectionName}".""")
        for variableName, li in mapCollection.items():
            if variableName in VARIABLE_ALIASES.keys():
                variableLookUpName = VARIABLE_ALIASES[variableName]
            else:
                variableLookUpName = variableName
            if variableLookUpName in matchedMaps0.keys():
                matchedMaps0[variableLookUpName].extend(li)
            else:
                matchedMaps0[variableLookUpName] = li
    matchedMaps = OrderedDict()
    for key in sorted(matchedMaps0.keys()):
        matchedMaps[key] = matchedMaps0[key]
    text = pprint.pformat(matchedMaps, indent=4)
    logging.info(f"""  The old and new maps have been matched to the variable names. Aliases have been used:\n{text}""")

    # Load, concatenate, and save maps by variable names
    logging.info("""Loading, concatenating, and saving maps by variable names.""")
    concatenatedMapsDict = {}
    for variableName, li in matchedMaps.items():
        logging.info(f"""  Working on variable "{variableName}".""")
        concatenatedMap = pd.DataFrame()
        emptyMap = makeMap(IDset=set(),
                           IDName=variableName,
                           startFrom=1,
                           irbNumber=IRB_NUMBER,
                           suffix=VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"],
                           columnSuffix=variableName,
                           deIdentificationMapStyle=DE_IDENTIFICATION_MAP_STYLE,
                           logger=logging.getLogger())
        newColumns = emptyMap.columns
        for fpath in li:
            fpath = Path(fpath)
            logging.info(f"""    Working on map located at "{fpath.absolute().relative_to(rootDirectory)}".""")
            df = pd.read_csv(fpath)
            oldColumns = df.columns
            logging.info(f"""  ..  Over-writing map header format to "{DE_IDENTIFICATION_MAP_STYLE}".""")
            logging.info(f"""  ..  Old map columns: {oldColumns.to_list()}.""")
            logging.info(f"""  ..  New map columns: {newColumns.to_list()}.""")
            df.columns = newColumns
            concatenatedMap = pd.concat([concatenatedMap, df])
        concatenatedMapsDict[variableName] = concatenatedMap
        fpath = runOutputDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]} map.csv")
        concatenatedMap.to_csv(fpath, index=False)

    # Quality control
    results = {}
    for variableName, df in concatenatedMapsDict.items():
        uniqueIDs = len(df.iloc[:, 0].unique())
        numIDs = len(df)
        percentDifference = getPercentDifference(uniqueIDs, numIDs)
        results[variableName] = {"Unique IDs": uniqueIDs,
                                 "Total IDs": numIDs,
                                 "Percent Similarity": percentDifference}
    resultsdf = pd.DataFrame.from_dict(results, orient="index")
    logging.info(f"Concatenation summary:\n{resultsdf.to_string()}")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
These functions convert columns by using conversion functions.
"""

import logging
import pprint
import sys
from pathlib import Path
from time import sleep
from typing import (Callable,
                    List,
                    Literal,
                    Union)
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (choosePathToLog,
                                    makeDirPath,
                                    makeMap,
                                    readDataFile)


def convertColumnsHash_byDir(listOfPortionDirs: list,
                             listOfPortionNames: List[str],
                             LIST_OF_PORTION_CONDITIONS: list,
                             columnsToConvert: list,
                             mapsColumnNames: dict,
                             logger: logging.Logger,
                             deIdentificationFunctions: dict,
                             VARIABLE_ALIASES: dict,
                             CHUNKSIZE: int,
                             runOutputDir: Path,
                             rootDirectory: Path,
                             MESSAGE_MODULO_CHUNKS: int,
                             SCRIPT_TEST_MODE: bool = False,):
    """
    """
    # Assert all portions have file criteria accounted for
    if len(listOfPortionDirs) == len(LIST_OF_PORTION_CONDITIONS):
        pass
    else:
        raise Exception("The number of portions does not equal the number of portion conditions.")

    # De-identify columns
    logger.info("""De-identifying files.""")
    allowLogging = False
    # Set file options: Map file
    for portionName, directory, fileConditions in zip(listOfPortionNames, listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        directory = Path(directory)  # For type hinting
        # Act on directory
        logger.info(f"""Working on directory "{choosePathToLog(directory, rootDirectory)}".""")
        listOfFiles = sorted(list(directory.iterdir()))
        for file in listOfFiles:
            logger.info(f"""  Working on file "{choosePathToLog(file, rootDirectory)}".""")
            fileOptions = {columnName: {"header": True,
                                        "mode": "w"} for columnName in columnsToConvert}
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options: Data file
                exportPath = runOutputDir.joinpath("Portions", portionName, f"{file.stem}.CSV")
                makeDirPath(exportPath.parent)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                logger.info("""  ..  Reading file to count the number of chunks.""")
                # >>> TEST BLOCK >>>  # TODO Remove
                if SCRIPT_TEST_MODE:
                    it1Total = 100
                else:
                    it1Total = sum([1 for _ in readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)])
                # <<< TEST BLOCK <<<
                logger.info(f"""  ..  This file has {it1Total:,} chunks.""")
                # Calculate logging requency
                if it1Total < MESSAGE_MODULO_CHUNKS:
                    moduloChunks = it1Total
                else:
                    moduloChunks = round(it1Total / MESSAGE_MODULO_CHUNKS)
                if it1Total / moduloChunks < 100:
                    moduloChunks = 1
                else:
                    pass
                dfChunks = readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)
                filePreview = readDataFile(file, nrows=2)
                filePreview = pd.DataFrame(filePreview)
                for it1, dfChunk0 in enumerate(dfChunks, start=1):
                    # >>> TEST BLOCK >>>  # TODO Remove
                    if SCRIPT_TEST_MODE:
                        if it1 > 1:
                            logger.info("""  ..  `SCRIPT_TEST_MODE` engaged. BREAKING.""")
                            break
                    # <<< TEST BLOCK <<<
                    dfChunk = pd.DataFrame(dfChunk0)
                    # Work on chunk
                    if it1 % moduloChunks == 0:
                        allowLogging = True
                    else:
                        allowLogging = False
                    if allowLogging:
                        logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        if columnName in VARIABLE_ALIASES.keys():
                            columnNameAlias = VARIABLE_ALIASES[columnName]
                            hasAlias = True
                        else:
                            columnNameAlias = columnName
                            hasAlias = False
                        # Logging block
                        if allowLogging:
                            if hasAlias:
                                logger.info(f"""  ..    Working on column "{columnName}" aliased as "{columnNameAlias}".""")
                            else:
                                logger.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in columnsToConvert or columnNameAlias in columnsToConvert:
                            if allowLogging:
                                logger.info("""  ..  ..  Column must be converted. Converting column...""")
                            # De-identify column: create de-identified values
                            newColumnName = mapsColumnNames[columnName]
                            newColumn = dfChunk[columnName].apply(deIdentificationFunctions[columnNameAlias])
                            newColumn = pd.Series(newColumn)  # For type hinting
                            newColumnWithOldName = newColumn
                            newColumnWithNewName = newColumn.rename(newColumnName)  # We can change `newColumnName` with the aliased version
                            # QA: Column/Variable Map: Variables
                            mapSavePath = runOutputDir.joinpath("Metadata", "Maps by Portion", portionName, f"{columnNameAlias}", f"{file.stem}.CSV")
                            makeDirPath(mapSavePath.parent)
                            mapHeader = fileOptions[columnNameAlias]["header"]
                            mapMode = fileOptions[columnNameAlias]["mode"]
                            # QA: Column/Variable Map: Save mapped column
                            chunkColumnMap = pd.concat([dfChunk[columnName], newColumnWithNewName], axis=1).drop_duplicates()
                            logger.info(f"""  ..  ..  Saving QA table to "{choosePathToLog(mapSavePath.absolute(), rootPath=rootDirectory)}".""")
                            logger.info(f"""  ..  ..  Saving QA table to `mapHeader` set to "{mapHeader}".""")
                            logger.info(f"""  ..  ..  Saving QA table to `mapMode` set to "{mapMode}".""")
                            chunkColumnMap.to_csv(mapSavePath, index=False, header=mapHeader, mode=mapMode)
                            fileOptions[columnNameAlias]["header"] = False
                            fileOptions[columnNameAlias]["mode"] = "a"
                            # De-identify column: Replace old values
                            dfChunk[columnName] = newColumnWithOldName
                            dfChunk = dfChunk.rename(columns={columnName: newColumnName})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileHeaders = False
                    fileMode = "a"
                    if allowLogging:
                        logger.info(f"""  ..  Chunk saved to "{choosePathToLog(exportPath, rootDirectory)}".""")
                # End file lopp
                pass
            else:
                logger.info("""    This file does not need to be processed.""")


def convertColumnsHash_byFile(file: Union[Path, str],
                              portionName: str,
                              columnsToConvert: list,
                              mapsColumnNames: dict,
                              deIdentificationFunctions: dict,
                              VARIABLE_ALIASES: dict,
                              CHUNKSIZE: int,
                              runOutputDir: Path,
                              rootDirectory: Path,
                              logger: logging.Logger,
                              MESSAGE_MODULO_CHUNKS: int,
                              SCRIPT_TEST_MODE: bool = False,):
    """
    """
    # Set file options: Data file
    file = Path(file)
    fileOptions = {columnName: {"header": True,
                                "mode": "w"} for columnName in columnsToConvert}
    exportPath = runOutputDir.joinpath("Portions", portionName, f"{file.stem}.CSV")
    makeDirPath(exportPath.parent)
    fileMode = "w"
    fileHeaders = True
    # Read file
    logger.info("""  ..  Reading file to count the number of chunks.""")
    # >>> TEST BLOCK >>>
    if SCRIPT_TEST_MODE:
        it1Total = 100
    else:
        it1Total = sum([1 for _ in readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)])
    # <<< TEST BLOCK <<<
    logger.info(f"""  ..  This file has {it1Total:,} chunks.""")
    # Calculate logging requency
    if it1Total < MESSAGE_MODULO_CHUNKS:
        moduloChunks = it1Total
    else:
        moduloChunks = round(it1Total / MESSAGE_MODULO_CHUNKS)
    if it1Total / moduloChunks < 100:
        moduloChunks = 1
    else:
        pass
    dfChunks = readDataFile(file, chunksize=CHUNKSIZE, low_memory=False)
    filePreview = readDataFile(file, nrows=2)
    filePreview = pd.DataFrame(filePreview)
    for it1, dfChunk0 in enumerate(dfChunks, start=1):
        # >>> TEST BLOCK >>>  # TODO Remove
        if SCRIPT_TEST_MODE:
            if it1 > 1:
                logger.info("""  ..  `SCRIPT_TEST_MODE` engaged. BREAKING.""")
                break
        # <<< TEST BLOCK <<<
        dfChunk = pd.DataFrame(dfChunk0)
        # Work on chunk
        if it1 % moduloChunks == 0:
            allowLogging = True
        else:
            allowLogging = False
        if allowLogging:
            logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
        for columnName in dfChunk.columns:
            # Work on column
            if columnName in VARIABLE_ALIASES.keys():
                columnNameAlias = VARIABLE_ALIASES[columnName]
                hasAlias = True
            else:
                columnNameAlias = columnName
                hasAlias = False
            # Logging block
            if allowLogging:
                if hasAlias:
                    logger.info(f"""  ..    Working on column "{columnName}" aliased as "{columnNameAlias}".""")
                else:
                    logger.info(f"""  ..    Working on column "{columnName}".""")
            if columnName in columnsToConvert or columnNameAlias in columnsToConvert:
                if allowLogging:
                    logger.info("""  ..  ..  Column must be converted. Converting column...""")
                # De-identify column: create de-identified values
                newColumnName = mapsColumnNames[columnName]
                newColumn = dfChunk[columnName].apply(deIdentificationFunctions[columnNameAlias])
                newColumn = pd.Series(newColumn)  # For type hinting
                newColumnWithOldName = newColumn
                newColumnWithNewName = newColumn.rename(newColumnName)  # We can change `newColumnName` with the aliased version
                # QA: Column/Variable Map: Variables
                mapSavePath = runOutputDir.joinpath("Metadata", "Maps by Portion", portionName, f"{columnNameAlias}", f"{file.stem}.CSV")
                makeDirPath(mapSavePath.parent)
                mapHeader = fileOptions[columnNameAlias]["header"]
                mapMode = fileOptions[columnNameAlias]["mode"]
                # QA: Column/Variable Map: Save mapped column
                chunkColumnMap = pd.concat([dfChunk[columnName], newColumnWithNewName], axis=1).drop_duplicates()
                logger.info(f"""  ..  ..  Saving QA table to "{choosePathToLog(mapSavePath.absolute(), rootPath=rootDirectory)}".""")
                logger.info(f"""  ..  ..  Saving QA table to `mapHeader` set to "{mapHeader}".""")
                logger.info(f"""  ..  ..  Saving QA table to `mapMode` set to "{mapMode}".""")
                chunkColumnMap.to_csv(mapSavePath, index=False, header=mapHeader, mode=mapMode)
                fileOptions[columnNameAlias]["header"] = False
                fileOptions[columnNameAlias]["mode"] = "a"
                # De-identify column: Replace old values
                dfChunk[columnName] = newColumnWithOldName
                dfChunk = dfChunk.rename(columns={columnName: newColumnName})
        # Save chunk
        dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
        fileHeaders = False
        fileMode = "a"
        if allowLogging:
            logger.info(f"""  ..  Chunk saved to "{choosePathToLog(exportPath, rootDirectory)}".""")
    # End file lopp
    pass


def convertColumnsHash(by: Literal["dir", "file"],
                       # Arguments used in preparation of case functions
                       IRB_NUMBER: str,
                       VARIABLE_SUFFIXES: dict,
                       # Arguments used in preparation AND passed to other functions: Common to both cases
                       VARIABLE_ALIASES: dict,
                       deIdentificationFunctions: dict,
                       logger: logging.Logger,
                       # Arguments passed to other functions: Common to both cases
                       rootDirectory: Path,
                       runOutputDir: Path,
                       CHUNKSIZE: int,
                       MESSAGE_MODULO_CHUNKS: int,
                       SCRIPT_TEST_MODE: bool,
                       # Arguments passed to other functions: "dir" case
                       listOfPortionDirs: list[Union[Path, str]] = None,
                       listOfPortionNames: list[str] = None,
                       LIST_OF_PORTION_CONDITIONS: list[Callable] = None,
                       # Arguments passed to other functions: "file" case
                       fpath: Path = None,
                       portionName: str = None
                       ) -> None:
    """
    """
    argsAsString = pprint.pformat(locals())
    logger.info(f"""`locals()`:\n{argsAsString}""")

    # Collect columns to convert
    columnsToConvert = sorted(list(deIdentificationFunctions.keys()))

    # Load de-identification maps for each variable that needs to be de-identified
    logger.info("""Loading de-identification maps for each variable that needs to be de-identified.""")
    mapsColumnNames = {}
    variablesCollected = [fname for fname in deIdentificationFunctions.keys()]
    for varName in variablesCollected:
        logger.info(f"""  Creating de-identified variable name for "{varName}".""")
        mapsColumnNames[varName] = f"De-identified {varName}"
    # Add aliases to `mapsColumnNames`
    for varName in VARIABLE_ALIASES.keys():
        varAlias = VARIABLE_ALIASES[varName]
        map_ = makeMap(IDset=set(),
                       IDName=varName,
                       startFrom=1,
                       irbNumber=IRB_NUMBER,
                       suffix=VARIABLE_SUFFIXES[varAlias]["deIdIDSuffix"],
                       columnSuffix=varName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        mapsColumnNames[varName] = map_.columns[-1]

    # QA: Test de-identification functions
    logger.info("""QA: Testing de-identification functions.""")
    for variableName, func in deIdentificationFunctions.items():
        logger.info(f"""  {variableName}: {func(1)}.""")

    # Run conversion
    if by == "dir":
        # Not implemented! TODO
        raise Exception("Sorry, this is not yet implemented! =(")
        convertColumnsHash_byDir(listOfPortionDirs=listOfPortionDirs,
                                 listOfPortionNames=listOfPortionNames,
                                 LIST_OF_PORTION_CONDITIONS=LIST_OF_PORTION_CONDITIONS,
                                 columnsToConvert=columnsToConvert,
                                 mapsColumnNames=mapsColumnNames,
                                 logger=logger,
                                 deIdentificationFunctions=deIdentificationFunctions,
                                 VARIABLE_ALIASES=VARIABLE_ALIASES,
                                 CHUNKSIZE=CHUNKSIZE,
                                 runOutputDir=runOutputDir,
                                 rootDirectory=rootDirectory,
                                 MESSAGE_MODULO_CHUNKS=MESSAGE_MODULO_CHUNKS,
                                 SCRIPT_TEST_MODE=SCRIPT_TEST_MODE)
    elif by in ["file", "file-dir"]:
        logger.info(f"""Working on file "{fpath}".""")
        convertColumnsHash_byFile(file=fpath,
                                  portionName=portionName,
                                  columnsToConvert=columnsToConvert,
                                  mapsColumnNames=mapsColumnNames,
                                  deIdentificationFunctions=deIdentificationFunctions,
                                  VARIABLE_ALIASES=VARIABLE_ALIASES,
                                  CHUNKSIZE=CHUNKSIZE,
                                  runOutputDir=runOutputDir,
                                  rootDirectory=rootDirectory,
                                  logger=logger,
                                  MESSAGE_MODULO_CHUNKS=MESSAGE_MODULO_CHUNKS,
                                  SCRIPT_TEST_MODE=SCRIPT_TEST_MODE)
        logger.info(f"""Working on file "{fpath}" - done.""")
    else:
        raise Exception(f"""`by` must be one of {{"dir", "file"}}, but instead got "{by}".""")



"""
Makes symbolic links
"""

import os
from pathlib import Path

LIST_OF_DIRECTORIES = ["Concatenated Results\data\output\makeMapsFromOthers\2023-11-07 12-05-06"]
LIST_OF_FILES = ["Concatenated Results\data\output\makeMapsFromScratch\2023-12-20 15-42-03\Patient Encounter Key map.csv"]
DESTINATION_FOLDER = "Concatenated Results\data\All Maps (SymLinks)"

destinationFolder = Path(DESTINATION_FOLDER)

for fpathString in LIST_OF_FILES:
    fpath = Path(fpathString)
    # SymLink to destination
    dest = destinationFolder.joinpath(fpath.name)
    os.symlink(fpath, dest)

for directoryString in LIST_OF_DIRECTORIES:
    directory = Path(directoryString)
    for fpath in directory.iterdir():
        # SymLink to DESTINATION
        dest = destinationFolder.joinpath(fpath.name)
        os.symlink(fpath, dest)




"""
QA: Make sure values were correctly converted.

# TODO Handle headers for aliased variables. Currently when tables are concatenated for the same variable but with different aliases, each alias is stored in a separate column. This can be done under `combineMaps`.
"""

import logging
from pathlib import Path
from typing_extensions import (Dict,
                               List,
                               Tuple,
                               Union)
# Third-party packages
import pandas as pd
from pandas.api.types import (is_numeric_dtype,
                              is_object_dtype)
# First-party packages
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getMapType)
from drapi.code.drapi.constants.constants import DE_IDENTIFICATION_PREFIXES


# Functions
def iterateOverFiles(listOfFiles: List[Union[str, Path]],
                     SCRIPT_TEST_MODE: bool,
                     logger: logging.Logger,
                     directoryName: Union[str, None] = None) -> pd.DataFrame:
    """
    """  # TODO Parallelize
    di = {}
    it1Total = len(listOfFiles)
    for it1, fpath in enumerate(listOfFiles, start=1):
        logger.info(f"""  Working on file {it1:,} of {it1Total:,}: "{fpath.name}".""")
        if SCRIPT_TEST_MODE:
            pass
        else:
            df = pd.read_csv(fpath, low_memory=False)
            result = getMapType(df=df)
            di[it1] = {"Directory": directoryName,
                       "File Name": fpath.name,
                       "Result": result}
    dfresult = pd.DataFrame.from_dict(data=di, orient="index")
    return dfresult


def qaListOfFilesAndDirectories(LIST_OF_FILES: List[Union[str, Path]],
                                LIST_OF_DIRECTORIES: List[Union[str, Path]],
                                SCRIPT_TEST_MODE: bool,
                                logger: logging.Logger):
    """
    """
    # Sort input alphabetically for easier debugging.
    if LIST_OF_FILES:
        listOfFiles1 = sorted([Path(fpath) for fpath in LIST_OF_FILES])
    else:
        listOfFiles1 = []
    if LIST_OF_DIRECTORIES:
        listOfDirectories = sorted([Path(dpath) for dpath in LIST_OF_DIRECTORIES])
    else:
        listOfDirectories = []

    # Work on files first
    dfresult1 = iterateOverFiles(listOfFiles=listOfFiles1,
                                 SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                 directoryName=None)

    # Work on directories
    resultsli = []
    for dpath in listOfDirectories:
        logger.info(f"""  Working on directory "{dpath.name}".""")
        listOfFiles2 = sorted(list(dpath.iterdir()))
        dfresult2 = iterateOverFiles(listOfFiles=listOfFiles2,
                                     SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                     directoryName=dpath.name)
        resultsli.append(dfresult2)

    dfresult = pd.concat([dfresult1] + resultsli, axis=0)
    return dfresult


def standardizeMapHeaders(dataFrame: pd.DataFrame,
                          variableAliases: Dict[str, str],
                          logger: logging.Logger) -> pd.DataFrame:
    """
    """
    TABLE_WIDTH = 2  # We assume the intermediate and final maps have only two columns.
    df = dataFrame
    columns = df.columns
    columnHeader1, columnHeader2 = columns
    if columnHeader1 in variableAliases.keys():
        identifiedVariableNameIndex = 0
        standardVariableName = variableAliases[columnHeader1]
    elif columnHeader2 in variableAliases.keys():
        identifiedVariableNameIndex = 1
        standardVariableName = variableAliases[columnHeader2]
    elif columnHeader1 in variableAliases.values():
        standardVariableName = columnHeader1
        identifiedVariableNameIndex = 0
    elif columnHeader2 in variableAliases.values():
        identifiedVariableNameIndex = 1
        standardVariableName = columnHeader2
    else:
        logger.warning(f"""Assuming variable is not aliased. Map columns were\n{columns}\nVariable aliases were\n{variableAliases}""")
        if any([columnHeader1.lower().startswith(prefix.lower()) for prefix in DE_IDENTIFICATION_PREFIXES.values()]):
            standardVariableName = columnHeader2
            identifiedVariableNameIndex = 1
        elif any([columnHeader2.lower().startswith(prefix.lower()) for prefix in DE_IDENTIFICATION_PREFIXES.values()]):
            standardVariableName = columnHeader1
            identifiedVariableNameIndex = 0
        else:
            raise Exception("This should not happen.")
    identifiedVariableName = columns[identifiedVariableNameIndex]
    # Standardize variable names: Deduce de-identified variable name location
    li = list(range(TABLE_WIDTH))
    li.pop(identifiedVariableNameIndex)
    deidentifiedVariableNameIndex = li[0]
    # Standardize variable names: Determine if de-identified variable contains standard or aliased name
    deidentifiedVariableName = columns[deidentifiedVariableNameIndex]
    deidentifiedVariableName = str(deidentifiedVariableName)
    deidentifiedVariableNameIsStandardized = standardVariableName in deidentifiedVariableName
    if deidentifiedVariableNameIsStandardized:
        newDeidentifiedVariableName = deidentifiedVariableName[:]
    else:
        # Determine which non-standard name is being used
        nameInAlias = False
        variableAliasesList = sorted(list(variableAliases.keys()),
                                     key=lambda string: len(string),
                                     reverse=True)
        for variableAlias in variableAliasesList:
            if variableAlias in deidentifiedVariableName:
                nameInAlias = variableAlias[:]
                break
            else:
                pass
        if not nameInAlias:
            if standardVariableName in deidentifiedVariableName:
                nameInAlias = standardVariableName
            else:
                raise Exception(f"""We were not able to deduce the identified version of the de-identified variable name.""")
        else:
            pass
        newDeidentifiedVariableName = deidentifiedVariableName.replace(nameInAlias,
                                                                       standardVariableName)
    # Standardize variable names: Rename variables
    df = df.rename(columns={identifiedVariableName: standardVariableName,
                            deidentifiedVariableName: newDeidentifiedVariableName})
    # Standardize order
    df = df[[standardVariableName, newDeidentifiedVariableName]]
    return df


def conformDataTypes(dataFrame: pd.DataFrame,
                     dataTypesDict: Dict[str, str],
                     logger=logging.Logger) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Converts dataframe datatypes according to what the IDR typically uses.
    """
    _ = logger
    df = dataFrame
    feedback = {}
    for it, variableName in enumerate(df.columns.to_list(), start=1):
        logger.info(f"""  Working on variable "{variableName}".""")
        if variableName in dataTypesDict.keys():
            # Lookup variable Python data type
            variableDataType0 = dataTypesDict[variableName]
            if variableDataType0.lower() == "numeric":
                variableDataType = "Int64"
            elif variableDataType0.lower() == "numeric_or_string":
                variableDataType = str
            elif variableDataType0.lower() == "string":
                variableDataType = str
            else:
                raise Exception("This should not happen")
            # Standardize Python data type # TODO We don't need this if the input data is clean, i.e., all values are of the same data type, e.g., there are no headers in the rows.
            if False:
                if variableDataType == str:
                    df[variableName] = df[variableName].astype(variableDataType)
                elif variableDataType == "Int64":
                    if is_numeric_dtype(df[variableName]):
                        df[variableName] = df[variableName].astype(variableDataType)
                    elif is_object_dtype(df[variableName]):
                        df[variableName] = df[variableName].astype(float).astype(variableDataType)
                    else:
                        raise Exception(f"""Encountered unexpected Pandas data type: "{df[variableName].dtypes}".""")
                else:
                    raise Exception(f"""Encountered unexpected data type: "{variableDataType}".""")
            elif True:
                dt1 = df[variableName].dtype
                logger.info(f"""    Converting data type from "{dt1}" to "{variableDataType}".""")
                df[variableName] = df[variableName].astype(variableDataType)
            result = True
        else:
            result = False
        feedback[it] = {"Column": variableName,
                        "Converted": result}
    feebackdf = pd.DataFrame.from_dict(data=feedback,
                                       orient="index")
    return df, feebackdf


def combineMaps(listOfDirectories: List[str],
                mode: int,
                variableAliases: Dict[str, str],
                dataTypesDict: Dict[str, str],
                runIntermediateDir: Path,
                runOutputDir: Path,
                rootPath: Path,
                rootDirectory: Path,
                logger: logging.Logger) -> List[Path]:
    """
    for `mode` = 1, assumes `listOfDirectories` is a tree with depth of 3, e.g. the "Maps by Portion" folder with the following structure:
    - Maps by Portion
        - Portion 1
            - Variable 1
                - CSV File 1
                - CSV File 2
                - etc.
            - Variable 2
                - CSV File 1
                - CSV File 2
                - etc.
            - etc.
        - Portion 2
            - Variable 1
                - CSV File 1
                - CSV File 2
                - etc.
            - Variable 2
                - CSV File 1
                - CSV File 2
                - etc.
            - etc.
        - etc.
    For `mode` = 2, assumes `listOfDirectories` is a tree with depth of 2.
    """
    listOfDirectories1 = sorted(list(listOfDirectories))
    dictOfVariables = {}
    it1 = 1
    for dpathString in listOfDirectories1:
        dpath = Path(dpathString)
        listOfVariables = sorted(list(dpath.iterdir()))
        for fpath1 in listOfVariables:
            variableDirName = fpath1.name
            if mode == 1:
                listOfFiles = sorted([fpath for fpath in fpath1.iterdir() if not fpath.name.startswith(".")])
                diNew = {it: {"Variable Name": variableDirName,
                              "File Path": fpath} for it, fpath in enumerate(listOfFiles, start=it1)}
            elif mode == 2:
                listOfFiles = [fpath1]
                diNew = {it: {"Variable Name": fpath.stem,
                              "File Path": fpath} for it, fpath in enumerate(listOfFiles, start=it1)}
            it1 += len(listOfFiles)
            dictOfVariables.update(diNew)

    dfAllPaths = pd.DataFrame.from_dict(data=dictOfVariables, orient="index")
    dfAllPaths.index.name = "Index"
    dfAllPaths = dfAllPaths.sort_values(by=["Variable Name", "File Path"])
    savePathAllPaths = runOutputDir.joinpath("Paths for All Maps.CSV")
    dfAllPaths.to_csv(path_or_buf=savePathAllPaths)

    # Combine maps by variable name
    listOfVariableNames = dfAllPaths["Variable Name"].drop_duplicates()
    it2Total = len(listOfVariableNames)
    savePathDir = runOutputDir.joinpath("Combined Maps", "By Variable")
    savePathDir.mkdir(parents=True)
    listOfPathsForCombinedMaps = []
    header = True
    for it2, variableName in enumerate(listOfVariableNames, start=1):
        logger.info(f"""  Working on variable {it2:,} of {it2Total:,}: "{variableName}".""")
        mask = dfAllPaths["Variable Name"] == variableName
        listOfPaths = dfAllPaths["File Path"][mask].to_list()
        it3Total = len(listOfPaths)
        listOfDataFramesPath = runIntermediateDir.joinpath(variableName)
        for it3, fpath in enumerate(listOfPaths, start=1):
            logger.info(f"""    Working on file {it3:,} of {it3Total:,}: "{choosePathToLog(path=fpath, rootPath=rootPath)}".""")
            df = pd.read_csv(filepath_or_buffer=fpath,
                             engine="pyarrow")
            # Standardize variable names
            df = standardizeMapHeaders(dataFrame=df,
                                       variableAliases=variableAliases,
                                       logger=logger)

            # TODO Remove this line, and fix upstream
            df = df.dropna(how="any")

            # Standardize Python data type
            df, _ = conformDataTypes(dataFrame=df,
                                     dataTypesDict=dataTypesDict,
                                     logger=logger)

            # Write results
            df.to_csv(path_or_buf=listOfDataFramesPath,
                      index=False,
                      mode="a",
                      header=header)
            header = False

        # Read file
        dfAllUnique = pd.read_csv(filepath_or_buffer=listOfDataFramesPath,
                                  low_memory=False)
        # Standardize Python data type
        df, _ = conformDataTypes(dataFrame=dfAllUnique,
                                 dataTypesDict=dataTypesDict,
                                 logger=logger)
        # Drop duplicates
        dfAllUnique = dfAllUnique.drop_duplicates()
        # Save results
        savePath = savePathDir.joinpath(f"{variableName}.CSV")
        logger.info(f"""    Saving combined maps to "{choosePathToLog(path=savePath, rootPath=rootDirectory)}".""")
        dfAllUnique.to_csv(path_or_buf=savePath,
                           index=False)
        listOfPathsForCombinedMaps.append(savePath)

    return listOfPathsForCombinedMaps



"""
Not implemented. For recent iterations see the following:
  - In the IDR shared drive: "/Volumes/FILES/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Shukla/IRB202001660 DatReq02/OMOP Portion/Code/getPersonIDs.py"
  - In this repo: "templates/makeDirTemplate/RenameMe OMOP Portion Template/Code/getPersonIDs.py
"""



"""
Script for modifying clinical text TSV files so they can be handled by the SDOH pipeline.
"""

import argparse
import logging
import pprint
from pathlib import Path
# Third-party packages
pass
# First-party packages
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.modifyTSV import modifyTSV


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("fpath",
                        help="The path(s) to the TSV file(s).",
                        type=str,
                        nargs="+")
    parser.add_argument("runFrom",
                        help="The directory to use as the anchor. There are two choices: `here` and `home`. Choosing `here` will use the present working directory as the anchor. Choosing `home` will use the DRAPI-Lemur install location as the anchor. The anchor is the point of reference for path construction. See the meta-parameters.",
                        choices=["here",
                                 "home"])

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")


    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    FPATHLI = argNamespace.fpath
    RUN_FROM = argNamespace.runFrom

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # Argument parsing: Additional checks
    pass

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    if RUN_FROM == "here":
        thisFilePath = Path(".")
    elif RUN_FROM == "home":
        thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Meta
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"
    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    itTotal = len(FPATHLI)
    for it, fpath0 in enumerate(FPATHLI, start=1):
        fpath = Path(fpath0)
        logger.info(f"""  Working on file {it:,} of {itTotal:,}: "{choosePathToLog(path=fpath, rootPath=rootDirectory)}".""")
        modifyTSV(fpath=fpath,
                  toDirectory=runOutputDir)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=rootDirectory)}".""")

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
QA: Make sure values were correctly converted.

# TODO Handle headers for aliased variables. Currently when tables are concatenated for the same variable but with different aliases, each alias is stored in a separate column. This can be done under `combineMaps`.
"""

import argparse
import json
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
pass
# Local packages
from drapi.code.drapi.constants.variableAliases import VARIABLE_ALIASES
from drapi.code.drapi.constants.constants import DATA_TYPES_DICT
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.parseAliasArguments import parseAliasArguments
from drapi.code.drapi.qa.deidentification import (combineMaps,
                                                  qaListOfFilesAndDirectories)


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("-d",
                        "--LIST_OF_DIRECTORIES",
                        nargs="*",
                        help="")
    parser.add_argument("-f",
                        "--LIST_OF_FILES",
                        nargs="*",
                        help="")

    parser.add_argument("-c",
                        "--COMBINE_MAPS",
                        help="""This is the path to the directory that contains the portions and their respective variables, typically "Metadata/Maps by Portion". When this option is used all maps are combined by variable name and the analysis is performed on the resulting tables. Input is a string that is converted to a Path object.""",
                        nargs="*",
                        type=Path)
    parser.add_argument("-m",
                        "--COMBINE_MAPS_MODE",
                        help="""Select "1" if combining atomic maps, select "2" if combining combined maps.""",
                        choices=[1, 2],
                        type=int)

    parser.add_argument("--CUSTOM_ALIASES",
                        help="""A JSON-formatted string of the form {`ALIAS`: `VARIABLE_NAME`}, where `VARIABLE_NAME` is the BO version of a variable name, and `ALIAS` is an alias of the variable name. An example is {"EncounterCSN": "Encounter # (CSN)"}.""",
                        type=json.loads)
    parser.add_argument("--DEFAULT_ALIASES",
                        help="""Indicates whether to include the default IDR aliases for variables.""",
                        action="store_true")

    parser.add_argument("-t",
                        "--SCRIPT_TEST_MODE",
                        help="Runs the script in a shorter fashion, to test its main features.",
                        action="store_true")

    # Arguments: Meta-variables
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    LIST_OF_DIRECTORIES = argNamespace.LIST_OF_DIRECTORIES
    LIST_OF_FILES = argNamespace.LIST_OF_FILES

    COMBINE_MAPS = argNamespace.COMBINE_MAPS
    COMBINE_MAPS_MODE = argNamespace.COMBINE_MAPS_MODE

    CUSTOM_ALIASES = argNamespace.CUSTOM_ALIASES
    DEFAULT_ALIASES = argNamespace.DEFAULT_ALIASES

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: Meta-variables
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    # Custom argument parsing: input paths
    if LIST_OF_FILES or LIST_OF_DIRECTORIES:
        pass
    elif COMBINE_MAPS:
        pass
    else:
        parser.error("Although `LIST_OF_FILES`, `LIST_OF_DIRECTORIES`, and `COMBINE_MAPS` are marked as optional in the help text, you must actually provide arguments for at least one of them.")

    # Custom argument parsing: aliases
    variableAliases = parseAliasArguments(customAliases=CUSTOM_ALIASES,
                                          useDefaultAliases=DEFAULT_ALIASES,
                                          defaultAliases=VARIABLE_ALIASES)
    # <<< Custom argument parsing <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USER_DOMAIN` = "{USER_DOMAIN}"
    `USERNAME` = "{USERNAME}"
    `USER_ID` = "{USER_ID}"
    `USER_PWD` = censored
    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin module body >>>
    if LIST_OF_FILES or LIST_OF_DIRECTORIES:
        dfresult = qaListOfFilesAndDirectories(LIST_OF_DIRECTORIES=LIST_OF_DIRECTORIES,
                                               LIST_OF_FILES=LIST_OF_FILES,
                                               SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                               logger=logger)
    elif COMBINE_MAPS:
        listOfPathsForCombinedMaps = combineMaps(listOfDirectories=COMBINE_MAPS,
                                                 mode=COMBINE_MAPS_MODE,
                                                 variableAliases=variableAliases,
                                                 dataTypesDict=DATA_TYPES_DICT,
                                                 runIntermediateDir=runIntermediateDir,
                                                 runOutputDir=runOutputDir,
                                                 rootPath=rootDirectory,
                                                 logger=logger)
        dfresult = qaListOfFilesAndDirectories(LIST_OF_DIRECTORIES=[],
                                               LIST_OF_FILES=listOfPathsForCombinedMaps,
                                               SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                               logger=logger)

    # Report results
    if SCRIPT_TEST_MODE:
        pass
    else:
        dfresult = dfresult.sort_values(by=["Directory",
                                            "File Name"])
        savePath = runOutputDir.joinpath("Map Type Results.CSV")
        dfresult.to_csv(savePath, index=False)
        dfString = dfresult.to_string()

        logger.info(f"""Map Type Results:\n{dfString}""")

    if not SCRIPT_TEST_MODE:
        logger.info(f"""Removing intermediate files.""")
        shutil.rmtree(runIntermediateDir)

    # <<< End module body <<<
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Creates an age de-identification map. This is an object that identifies patients who are aged 90 or more at the time of processing (i.e., now, when the script is run).

This is usually run before `deIdentifyByAge`.

Optionally this also creates a second map using a second set of ID's, which is usually necessary if you are operating on a data set that has already had its IDs de-identified. This optional function uses the following parameters:

    - `SECONDARY_MAP`
    - `SECONDARY_MAP_COLUMN_NAME_FROM`
    - `SECONDARY_MAP_COLUMN_NAME_TO`

Note that there must be a variable (column) in common between the tables `SECONDARY_MAP` and `INPUT_FILE`, for the patients to be correctly mapped from one table to the other.
"""

import logging
import os
from datetime import datetime as dt
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
INPUT_FILE = Path(r"..\Concatenated Results\data\output\convertColumns\...\person.csv")
PATIENT_ID_NAME = "Patient Key"
DATE_OF_BIRTH_COLUMN = "birth_datetime"
AGE_THRESHOLD = 90

SECONDARY_MAP = Path(r"..\Concatenated Results\data\output\concatenateMaps\...\Patient Key map.csv")  # Path object or `None`
SECONDARY_MAP_COLUMN_NAME_FROM = "Patient Key"
SECONDARY_MAP_COLUMN_NAME_TO = "De-identified Patient Key"

# Arguments; General
CHUNK_SIZE = 50000

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `INPUT_FILE`: "{INPUT_FILE}"
    `PATIENT_ID_NAME`: "{PATIENT_ID_NAME}"
    `DATE_OF_BIRTH_COLUMN`: "{DATE_OF_BIRTH_COLUMN}"
    `AGE_THRESHOLD`: "{AGE_THRESHOLD}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    logger.info(f"""Working on file "{INPUT_FILE.absolute().relative_to(rootDirectory)}".""")
    # Set file options
    exportPath = runIntermediateDataDir.joinpath(INPUT_FILE.name)
    fileMode = "w"
    fileHeaders = True
    # Read file
    logger.info("""    Reading file to count the number of chunks.""")
    numChunks = sum([1 for _ in pd.read_csv(INPUT_FILE, chunksize=CHUNK_SIZE)])
    logger.info(f"""    This file has {numChunks} chunks.""")
    dfChunks = pd.read_csv(INPUT_FILE, chunksize=CHUNK_SIZE)
    for it, dfChunk in enumerate(dfChunks, start=1):
        dfChunk = pd.DataFrame(dfChunk)
        # Work on chunk
        logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
        dfChunk = dfChunk[[PATIENT_ID_NAME, DATE_OF_BIRTH_COLUMN]]
        dfChunk = dfChunk.drop_duplicates()
        dfChunk = dfChunk.set_index(PATIENT_ID_NAME)
        series = dfChunk[DATE_OF_BIRTH_COLUMN]
        series = pd.to_datetime(series)
        currentAge = dt.today() - series
        currentAge = currentAge.apply(lambda timeDelta: timeDelta.days / 365.25)
        dfChunk["Over Age Threshold"] = currentAge >= AGE_THRESHOLD
        dfChunk["Over Age Threshold"] = dfChunk["Over Age Threshold"].apply(lambda boolean: 1 if boolean else 0)

        # Save chunk
        dfChunk["Over Age Threshold"].to_csv(exportPath, mode=fileMode, header=fileHeaders, index=True)
        fileMode = "a"
        fileHeaders = False
        logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")

    # Sort and remove duplicates from map
    logger.info("Sorting and removing duplicates from map.")
    df1 = pd.read_csv(exportPath)
    df1 = df1.drop_duplicates()
    df1 = df1.sort_values(by=[PATIENT_ID_NAME, "Over Age Threshold"])
    logger.info("Sorting and removing duplicates from map - done.")
    savePath1 = runOutputDir.joinpath(f"Age Map - {PATIENT_ID_NAME}.CSV")
    metadataText = f"""# Map generation parameters: `AGE_THRESHOLD`: "{AGE_THRESHOLD}".
# Map generation parameters: `PATIENT_ID_NAME`: "{PATIENT_ID_NAME}".
# Map generation parameters: `DATE_OF_BIRTH_COLUMN`: "{DATE_OF_BIRTH_COLUMN}".
"""
    with open(savePath1, "w") as file:
        file.write(metadataText)
    df1.to_csv(savePath1, index=False, mode="a")

    # Create a map using a secondary map
    if SECONDARY_MAP:
        df1 = df1.set_index(PATIENT_ID_NAME)
        df2 = pd.read_csv(SECONDARY_MAP)
        df2 = df2.set_index(SECONDARY_MAP_COLUMN_NAME_FROM)
        newMap = df1.join(other=df2[SECONDARY_MAP_COLUMN_NAME_TO],
                          how="inner")
        newMap = newMap[[SECONDARY_MAP_COLUMN_NAME_TO, "Over Age Threshold"]]
        savePath2 = runOutputDir.joinpath(f"Age Map - {SECONDARY_MAP_COLUMN_NAME_TO}.CSV")
        with open(savePath2, "w") as file:
            file.write(metadataText)
        newMap.to_csv(savePath2, index=False, mode="a")
    else:
        pass

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
Utility program to visualize the structure of a directory and its contents in the form of a tree. Compare this to the Linux built-in `tree`.
"""

USAGE_INSTRUCTIONS = """Usage instructions

To use this program, just enter in the command line:

`python tree.py directory_path`

where `directory_path` is the path of the directory you want the tree of. If you are in the directory you want the tree for, then you just do

`python tree.py .`

Note the above example assumes "tree.py" is in the same directory that you are operating on. If the program is in another directory, you must type its path as below

`python path/to/tree.py directory_path`
"""

# NOTE The function `tree` is copied from, and is also available in the `drapi` module.

import argparse
import sys
# Package imports
from drapi.code.drapi.drapi import tree


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("dir_path",
                        help="The path of the directory for which to get the tree.",
                        type=str)

    parser.add_argument("--level",
                        help="The level.",
                        default=-1,
                        type=int)

    parser.add_argument("--limit_to_directories",
                        help="Limit to directories.",
                        default=False,
                        type=bool)

    parser.add_argument("--length_limit",
                        help="The length limit.",
                        default=1000,
                        type=int)

    if not len(sys.argv) > 1:
        parser.print_usage()
        print(USAGE_INSTRUCTIONS)
        sys.exit(0)
    else:
        args = parser.parse_args()

    dir_path = args.dir_path
    level = args.level
    limit_to_directories = args.limit_to_directories
    length_limit = args.length_limit

    tree(dir_path=dir_path,
         level=level,
         limit_to_directories=limit_to_directories,
         length_limit=length_limit)

